//=============================================================================
//  Unified IO - C++ interface for industrial IO cards
//  Copyright (C) 2000, UAB "BBD SOFT" ( http://www.bbdsoft.com/ )
//
// This material is provided "as is", with absolutely no warranty expressed
// or implied. Any use is at your own risk.
//
// Permission to use or copy this software for any purpose is hereby granted 
// without fee, provided the above notices are retained on all copies.
// Permission to modify the code and to distribute modified code is granted,
// provided the above notices are retained, and a notice that the code was
// modified is included with the above copyright notice.
//
//  The author of this program may be contacted at developers@bbdsoft.com
//=============================================================================


#ifndef _BTRACE_
   #include "btrace.hpp"
#endif

#ifndef _GEMICROGENI_
   #include "gemicrogeni.hpp"
#endif

#ifndef _STDEXCEPT_
   #include <stdexcept>
#endif

#ifndef _IOPORT_
   #include "ioport.hpp"
#endif

#ifndef _IOSTREAM_
   #include <iostream>
#endif

#ifndef _IOTRACK_
   #include "iotrack.hpp"
#endif

#ifndef _OSIFACE_
   #include "osiface.hpp"
#endif

using namespace std;
using namespace Uniio;


//-----------------------------------------------------------------------------
GEMicroGENI::GEMicroGENI ( const unsigned inIOPortAddress
                         , const unsigned long inSharedRamAddress )
   : IOCard ("GE-MicroGENI")
   , theSharedMemory ( inSharedRamAddress , 0x4000, true )
   , theBaseIOAddress (inIOPortAddress)
   , theSizeOfIOTable (0)
{
BFUNCTRACE_DEVELOP ();
   if (!isCommOK() || !isBoardOK())
   {
      clearReset ();
      reset (false);
      OSInterface::osinterface().sleep (10000);
      reset ();
      clearReset ();
   }
   for ( int i=0
       ; (i < 5000) && (!isCommOK() || !isBoardOK())
       ; i++ )
   {
      OSInterface::osinterface().sleep (10000);
   }
   if (!isCommOK() || !isBoardOK())
   {
      runtime_error exc ("GEMicroGENI: board failed to initialize");
      throw exc;
   }
   getSizeOfIOTable();
} // end constructor


//-----------------------------------------------------------------------------
GEMicroGENI::~GEMicroGENI ()
{
BFUNCTRACE_DEVELOP ();
} // end destructor


//-----------------------------------------------------------------------------
bool GEMicroGENI::isBoardOK () const
{
BFUNCTRACE_DEVELOP ();
  IOPort statusRegister ( baseIOAddress()
                          , false               // write access not required
                          );
  unsigned char status = statusRegister.readChar ();
return !(status & 0x8);
} // end isBoardOK (..)

//-----------------------------------------------------------------------------
bool GEMicroGENI::isCommOK () const
{
BFUNCTRACE_DEVELOP ();

IOPort statusRegister ( baseIOAddress()
                      , false               // write access not required
                      );
unsigned char status = statusRegister.readChar ();
return !(status & 0x10);
} // end GEMicroGENI::isCommOK (..) const


//-----------------------------------------------------------------------------
bool GEMicroGENI::isReset () const
{
BFUNCTRACE_DEVELOP ();

IOPort statusRegister ( baseIOAddress()
                      , false               // write access not required
                      );
unsigned char status = statusRegister.readChar ();

return !(status & 0x1);
} // end GEMicroGENI::isResetOK (..) const


//-----------------------------------------------------------------------------
GEMicroGENI & GEMicroGENI::reset (const bool inReset)
{
BFUNCTRACE_DEVELOP ();

IOPort controlRegister ( baseIOAddress()+1);
unsigned char aByte = controlRegister.readChar();

if (!inReset)
{
   aByte &= ~0x40;
}
else
{
   aByte |= 0x40;
} // endifelse

controlRegister.writeChar (aByte);
return *this;
} // end GEMicroGENI::reset (..)


//-----------------------------------------------------------------------------
GEMicroGENI & GEMicroGENI::clearReset ()
{
BFUNCTRACE_DEVELOP ();

IOPort controlRegister ( baseIOAddress()+1);
unsigned char aByte = controlRegister.readChar();
aByte &= ~0x2;
controlRegister.writeChar (aByte);
aByte |= 0x2;
controlRegister.writeChar (aByte);
return *this;
} // end GEMicroGENI::clearReset (..)


//-----------------------------------------------------------------------------
IOCard & GEMicroGENI::updateInputs ( const IOTrack & inTrack
                                   , long * inImage
                                   , long & inTrackState
                                   )
{
BFUNCTRACE_DEVELOP ();

if (!isCommOK() || !isBoardOK())
{
   inTrackState = IOTrack::interfaceBoardFault;
}
else
{
   if (!stationExists(inTrack.startingPoint()))
   {
      inTrackState = IOTrack::moduleFault;
   }
   else
   {
      inTrackState = IOTrack::noFault;
   } // endifelse
} // endifelse

if ( inTrack.hasPointOfType ('I'))
{
   readDigitalInputs ( inTrack , inImage, inTrackState );
}
else
{
   if (inTrack.hasPointOfType ('A'))
   {
      readAnalogInputs ( inTrack , inImage, inTrackState );
   } // endif
} // endifelse
return * this;
} // end GEMicroGENI::updateInputs (..)


//-----------------------------------------------------------------------------
void GEMicroGENI::readDigitalInputs ( const IOTrack & inTrack
                                  , long * inImage
                                  , long & inTrackState
                                  )
{
BFUNCTRACE_DEVELOP ();

for ( int i = 0
    ; i < inTrack.numberOfPoints()/pointsPerBank
    ; i++)
{
   // to increase processing speed update only banks that have at least one input
   if ( ~theBankMasks[i+(inTrack.startingPoint()*4)] )
   {
      unsigned char newValue = theSharedMemory.readChar ( IOTableOffset
                                      + (sizeOfIOTable()*inTrack.startingPoint())
                                      + i
                                      );

      int imageBase = i*pointsPerBank;
      for (int j = 0; j<pointsPerBank; j++)
      {
         if ( (~theBankMasks[i+(inTrack.startingPoint()*4)]) & (1 << j))
         {
            inImage[imageBase+j] = newValue %2;
         }
         newValue /= 2;
      } // endfor
   } // endif
} // endfor

} // end GEMicroGENI::readDigitalInputs (..)


//-----------------------------------------------------------------------------
void GEMicroGENI::readAnalogInputs ( const IOTrack & inTrack
                                   , long * inImage
                                   , long & inTrackState
                                   )
{
BFUNCTRACE_DEVELOP ();

   for ( int i = 0
       ; i < inTrack.numberOfPoints()
       ; i++)
   {
      if (inTrack.pointType( i ) == IOTrack::analogInput )
      {
         // read and convert 16 bit analog value from Genius to 32 bit value
         unsigned short newValue = theSharedMemory.readShort ( IOTableOffset
                                      + (sizeOfIOTable()*inTrack.startingPoint())
                                      + i
                                      );
         inImage [i] = (newValue << 16) - 0x80000000 ;
      } // endif
   } // endfor

} // end GEMicroGENI::readAnalogInputs (..)


//-----------------------------------------------------------------------------
void GEMicroGENI::writeAnalogOutputs ( const IOTrack & inTrack
                                     , const long * const inImage
                                     , long & inTrackState
                                     )
{
BFUNCTRACE_DEVELOP ();

for ( int i = 0
    ; i < inTrack.numberOfPoints()
    ; i++)
{
   if (inTrack.pointType( i ) == IOTrack::analogOutput )
   {
      unsigned short newValue = (inImage [i] >> 16) & 0xFFFF;
      // read and convert 16 bit analog value from Genius to 32 bit value
      theSharedMemory.writeShort ( IOTableOffset
                                 + (sizeOfIOTable()*inTrack.startingPoint())
                                 + i
                                 , newValue
                                 );
   } // endif
} // endfor

} // end GEMicroGENI::writeAnalogOutputs (..)


//-----------------------------------------------------------------------------
IOCard & GEMicroGENI::updateOutputs ( const IOTrack & inTrack
                                    , const long * const inImage
                                    , long & inTrackState
                                    )
{
BFUNCTRACE_DEVELOP ();

if (!isCommOK() || !isBoardOK())
{
   inTrackState = IOTrack::interfaceBoardFault;
}
else
{
   if (!stationExists(inTrack.startingPoint()))
   {
      inTrackState = IOTrack::moduleFault;
   }
   else
   {
      inTrackState = IOTrack::noFault;
   } // endifelse
} // endifelse

if (inTrack.hasPointOfType ('O'))
{
   writeDigitalOutputs ( inTrack , inImage, inTrackState );
}
else
{
   if (inTrack.hasPointOfType ('U'))
   {
      writeAnalogOutputs ( inTrack , inImage, inTrackState );
   } // endif
} // endifelse
return * this;
} // end GEMicroGENI::updateOutputs (..)


//-----------------------------------------------------------------------------
void GEMicroGENI::writeDigitalOutputs ( const IOTrack & inTrack
                                   , const long * const inImage
                                   , long & inTrackState
                                   )
{
BFUNCTRACE_DEVELOP ();


for ( int i = 0
    ; i < inTrack.numberOfPoints()/pointsPerBank
    ; i++)
{
   // to increase processing speed update only banks that have at least one output
   if ( theBankMasks[i+(inTrack.startingPoint()*4)] )
   {
      unsigned char newValue = 0;

      int imageBase = i*pointsPerBank;

      for (int j = 0; j<pointsPerBank; j++)
      {
         if ( inImage[imageBase+j] )
         {
            newValue |= (1 << j);
         } // endif
      } // endfor

      // this is needed to prevent writing 1 to not used points
      newValue &= theBankMasks[i+(inTrack.startingPoint()*4)];

      theSharedMemory.writeChar ( IOTableOffset
                    + ((maxStations+inTrack.startingPoint())*sizeOfIOTable())
                    + i
                    , newValue
                    );
   } // endif
} // endfor
} // end GEMicroGENI::updateOutputs (..)


//-----------------------------------------------------------------------------
IOCard & GEMicroGENI::addTrack ( IOTrack & inTrack )
{
BFUNCTRACE_DEVELOP ();

if (   inTrack.hasPointOfType ('C') 
    || inTrack.hasPointOfType ('R') 
   )
{
   throw logic_error ("GEMicroGENI: Current implementation does not support counters.");
} // endif

if (  (inTrack.hasPointOfType ('O') || inTrack.hasPointOfType ('I'))
   && (  (inTrack.numberOfPoints() % pointsPerBank)
      || (inTrack.numberOfPoints() > 32)
      )
   )
{
   logic_error exc ("GEMicroGENI: Digital IO track length can only by n*8 points but no more than 32");
   throw exc;
} // endif

if ( inTrack.startingPoint() > maxStations-1)
{
   throw range_error ("GEMicroGENI: starting point parameter value is too high. GEMicroGENI supports up to 32 stations only.");
} // endif

   if (stationExists(inTrack.startingPoint()))
   {
      unsigned char anIOConfig = theSharedMemory.readChar ( DeviceConfigurationTableOffset
                                   + inTrack.startingPoint()*8
                                   + 7
                                   );
      switch (anIOConfig)
      {
         case 1:
         {
            // All inputs
            if (  inTrack.hasPointOfType ('O')
               || inTrack.hasPointOfType ('U'))
            {
               throw range_error ("GEMicroGENI: station is configured as all Inputs.");
            }
            break;
         }
         case 2:
         {
            // All outputs
            if (  inTrack.hasPointOfType ('I')
               || inTrack.hasPointOfType ('A'))
            {
               throw range_error ("GEMicroGENI: station is configured as all Outputs.");
            }
            break;
         }
         case 3:
         {
            // mix
            break;
         }
         default:
         {
            throw range_error ("GEMicroGENI: station is not configured.");
         }
      } // end switch
   } // end if

bool fOutputs = false;
if (  inTrack.hasPointOfType ('O') || inTrack.hasPointOfType ('I'))
{
   if (  inTrack.hasPointOfType ('U')
      || inTrack.hasPointOfType ('A'))
   {
      throw range_error ("GEMicroGENI: Genius station can not be analog and digital mix.");
   }
   // initialize all required banks
   for ( int i = 0
       ; i < inTrack.numberOfPoints()/pointsPerBank
       ; i++)
   {
      // pack IO mask for 8 points into one byte
      theBankMasks[i+(inTrack.startingPoint()*4)]= 0;
      for (int j = 0; j < pointsPerBank; j++)
      {
         if ( inTrack.pointType( i * pointsPerBank + j) == IOTrack::digitalOutput )
         {
            theBankMasks[i+(inTrack.startingPoint()*4)] |= (1 << j);
            fOutputs = true;
         } // endif
      } // endfor
   } // endfor
} // end if digital
else
{
   fOutputs = inTrack.hasPointOfType (IOTrack::analogOutput);
}
if (fOutputs)
{
      // enable changing outputs
   theSharedMemory.writeChar ( DeviceConfigurationTableOffset
                                   + inTrack.startingPoint()*8
                                   + 1
                                   , 0
                                   );

}

return * this;
} // end GEMicroGENI::addTrack (..)


//-----------------------------------------------------------------------------
void GEMicroGENI::getSizeOfIOTable()
{
BFUNCTRACE_DEVELOP ();

unsigned char aSize = theSharedMemory.readChar (SetupTableOffset+5);
if ((aSize == 0) || (aSize > 128))
{
   throw range_error ("GEMicroGENI: size of IO Table is invalid.");
} // endif
theSizeOfIOTable = aSize;
} // end GEMicroGENI::getSizeOfIOTable () 


//-----------------------------------------------------------------------------
bool GEMicroGENI::stationExists (const unsigned inStation) const
{
BFUNCTRACE_DEVELOP ();
BASSERTTEXT (inStation < maxStations, "GEMicroGENI supports up to 32 stations");

char aPresent = theSharedMemory.readChar ( DeviceConfigurationTableOffset
                                   + inStation*8
                                   + 2
                                   );
return (aPresent == 1);
} // GEMicroGENI::stationExists (..) const


//-----------------------------------------------------------------------------
GEMicroGENI & GEMicroGENI::lockout() 
{
BFUNCTRACE_DEVELOP ();
unsigned char lockoutState;
unsigned char lockoutRequest;

do
{
   lockoutState = theSharedMemory.readChar ( LockoutStateOffset );
   lockoutRequest = theSharedMemory.readChar ( LockoutRequestOffset );
} while (lockoutState != 0 && lockoutRequest !=0);

lockoutRequest = 1;

   theSharedMemory.writeChar ( LockoutRequestOffset, lockoutRequest);
do
{
   lockoutState = theSharedMemory.readChar ( LockoutStateOffset );
} while (lockoutState == 0);

return *this;
} // end GEMicroGENI::lockout ()


//-----------------------------------------------------------------------------
GEMicroGENI & GEMicroGENI::release() 
{
BFUNCTRACE_DEVELOP ();
unsigned char lockoutState;
unsigned char lockoutRequest;

   lockoutState = theSharedMemory.readChar ( LockoutStateOffset );
   lockoutRequest = theSharedMemory.readChar ( LockoutRequestOffset );

if (lockoutState == 1 && lockoutRequest == 1);
{

   lockoutRequest = 0;
   theSharedMemory.writeChar ( LockoutRequestOffset, lockoutRequest);
   do
   {
      lockoutState = theSharedMemory.readChar ( LockoutStateOffset );
   } while (lockoutState == 1);
} // endif
return *this;
} // end GEMicroGENI::release ()

